/* $Id: sib.c,v 1.3 2000/06/04 20:22:56 nop Exp $
 *
 * SIB - Routines for Serial Interconnect Bus on r39xx
 *       (used for sound, telecom, touchscreen, and other interfaces)
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2000 Jim Pick <jim@jimpick.com>
 */

#include <linux/wait.h>
#include <linux/sched.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/r39xx.h>
#include <asm/betty.h>
#include "sib.h"

#ifdef CONFIG_VTECH_HELIO_EMULATOR

/*  
 * The emulator is horribly confused by all this high performance
 * interrupt-driven code.  Fortunately, it doesn't need any delays to
 * run anything either, so a very simple version of read/write for the
 * sib follows.
 */

void sib_write_check(int reg, unsigned int data) 
{
        unsigned int d;

        d = data | SIB_WRITE;
        d = d | ((reg << SIB_ADDRESS_SHIFT) & SIB_ADDRESS_MASK);

        SIBSubFrame0Control = d;
}

void sib_read_check(int reg, void (*next_read_callback)(void))
{
        unsigned int d;

        d = ((reg << SIB_ADDRESS_SHIFT) & SIB_ADDRESS_MASK);

        SIBSubFrame0Control = d;

        next_read_callback();
}

void sib_init(void) 
{
  IntEnable1 &= ~(INT1_SIBSF0INT | INT1_SIBSF1INT);
#if 0
	PowerControl |= PWR_DIVMOD;  /* Peripheral clock speeds are
                                        independent from CPU speed */


	SIBControl &= ~SIB_SIBSCLKDIV_MASK;

	ClockControl &= ~CLK_SIBMCLKDIV_MASK;
        /* This forces SIBSF0 interrupts if it's on, even if the interrupts are disabled. */
#if 0
	ClockControl |= CLK_SIBMCLKDIR | CLK_ENSIBMCLK |
		(2 << CLK_SIBMCLKDIV_SHIFT) | CLK_CSERSEL;
#endif

	SIBControl &= ~(SIB_ENSIB | SIB_ENSF0);
#endif
}

#else


#define SIB_WRITEBUFFER_SIZE   40
static struct sib_write_struct writebuffer[ SIB_WRITEBUFFER_SIZE ];
static struct sib_write_struct *writebuffer_head = writebuffer; /* Write here */
static struct sib_write_struct  *writebuffer_tail = writebuffer; /* Read here */
static int writebuffer_count = 0; /* Number of writes in buffer */
static void (*next_read_callback)(void) = NULL;


/*
 *  Interrupt routine
 */

static void sib_interrupt_sf0(int irq, void *dev_id,
					 struct pt_regs * regs)
{
	static int counter = 0;

	IntClear1 = INT1_SIBSF0INT;  

	if (next_read_callback != NULL) {
/*		printk("next_read_callback: %p\n", next_read_callback); */
		next_read_callback();
	}


	if (writebuffer_count == 0) {
/*
		if (counter++ < 5)
		  printk(" %08x %08x\n", IntStatus1, IntEnable1 );
*/
	       	SIBControl &= ~(SIB_ENSIB | SIB_ENSF0);
		IntEnable1 &= ~INT1_SIBSF0INT;
		IntClear1 = INT1_SIBSF0INT;
		next_read_callback = NULL;
	}
	else {
/*		printk( "Out: 0x%08x\n", writebuffer_tail->data ); */
		SIBSubFrame0Control = writebuffer_tail->data;
		next_read_callback = writebuffer_tail->next_read_callback;
		writebuffer_tail++;
		if (writebuffer_tail == writebuffer + SIB_WRITEBUFFER_SIZE)
			writebuffer_tail = writebuffer;
		writebuffer_count--;
/*
		if (writebuffer_count == SIB_WRITEBUFFER_SIZE - 1)
			wake_up_interruptible(&sib_write_wait);
*/
	}
}


/*
 *  Initialization routine
 */

void sib_init(void)
{
	int i;

#ifdef CONFIG_VTECH_HELIO
	/* Reset Betty (wired to IR RXPWR pin?) */

	IRControl1 |= IR_RXPWR;      /* On */
	for (i = 0; i < 16; i++);

	IRControl1 &= ~IR_RXPWR;     /* Off */
	for (i = 0; i < 256; i++);

	IRControl1 |= IR_RXPWR;      /* On */
	for (i = 0; i < 16; i++);
#endif

	PowerControl |= PWR_DIVMOD;  /* Peripheral clock speeds are
                                        independent from CPU speed */


	SIBControl &= ~SIB_SIBSCLKDIV_MASK;

	ClockControl &= ~CLK_SIBMCLKDIV_MASK;
	ClockControl |= CLK_SIBMCLKDIR | CLK_ENSIBMCLK |
		(2 << CLK_SIBMCLKDIV_SHIFT) | CLK_CSERSEL;

	SIBControl &= ~(SIB_ENSIB | SIB_ENSF0);
	IntEnable1 &= ~INT1_SIBSF0INT;
	IntClear1 = INT1_SIBSF0INT;

	if (i = request_irq(40, sib_interrupt_sf0, SA_INTERRUPT,
			"sib sf0", NULL)) {
		printk(KERN_ERR "sib: Cannot allocate irq for SF0. (%d)\n",
		       i);
	}

 
}


/*
 *  Send data on SIB in subframe 0 (subframe 1 is unused?)
 */

static int sib_command(int reg, unsigned int data, void (*next_read_callback)(void))
{
        /* printk("C: %d %d\n", writebuffer_count, SIB_WRITEBUFFER_SIZE); */

	/* If the buffer is full, then return failure */
	if (writebuffer_count == SIB_WRITEBUFFER_SIZE)
		return -1;

	writebuffer_head->data = data | 
		((reg << SIB_ADDRESS_SHIFT) & SIB_ADDRESS_MASK);
	writebuffer_head->next_read_callback = next_read_callback;
	writebuffer_head++;
	if (writebuffer_head == writebuffer + SIB_WRITEBUFFER_SIZE)
		writebuffer_head = writebuffer;
	writebuffer_count++;
	SIBControl |= SIB_ENSIB | SIB_ENSF0;
	IntEnable1 |= INT1_SIBSF0INT;
	IntClear1 = INT1_SIBSF0INT;

	return 0;
}

/*
 *  Send data on SIB in subframe 0 (subframe 1 is unused?)
 */

int sib_write(int reg, unsigned int data)
{
/*	printk( "W: %d, 0x%04x\n", reg, data ); */
	return sib_command( reg, data | SIB_WRITE, NULL );
}

void sib_write_check(int reg, unsigned int data)
{
	if (sib_write( reg, data ))
		printk("SIB buffer full - command lost.\n");
}


/*
 *  Send data on SIB in subframe 0 and wait for returned data
 */

int sib_write_read(int reg, unsigned int data,
			      void (*next_read_callback)(void))
{
/*	printk( "WR: %d, 0x%04x %p\n", reg, data, next_read_callback ); */
	return sib_command( reg, data | SIB_WRITE, next_read_callback );
}

int sib_read(int reg, void (*next_read_callback)(void))
{
/*	printk( "R: %d %p\n", reg, next_read_callback ); */
	return sib_command( reg, 0, next_read_callback );
}

void sib_read_check(int reg, void (*next_read_callback)(void))
{
	if (sib_read( reg, next_read_callback ))
		printk("SIB buffer full - command lost.\n");
}

#endif
