/*
 * linux/arch/mips/vr41xx/int-handler.S
 *
 * Copyright (C) 1995, 1996, 1997 Paul M. Antoine and Harald Koerfgen
 *
 * Written by Ralf Baechle and Andreas Busse, modified for DECStation
 * support by Paul Antoine and Harald Koerfgen.
 *
 * completly rewritten:
 * Copyright (C) 1998 Harald Koerfgen
 *
 * Adapted to the VR41XX and almost entirely rewritten:
 * Copyright (C) 1999 Bradley D. LaRonde and Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <asm/asm.h>
#include <asm/regdef.h>
#include <asm/mipsregs.h>
#include <asm/stackframe.h>
#include <asm/vr41xx.h>


		.text
		.set	noreorder

/*
 * Interrupt handler for VR41xx based devices
 *
 * In a zealous attempt to save CPU cycles, we don't worry about making the code
 * readable and instead optimize for the most frequently occurring ints (the
 * timer int IRQ 3 and the ICU ints that cascade off IRQ 2), while keeping a
 * priority scheme that actually makes some sense:
 *
 * Highest ----     CPU IP3 - Int1: RTClong1    IRQ  3
 *                  CPU IP4 - Int2: RTClong2    IRQ  4
 *                  CPU IP5 - Int3: HSP         IRQ  5
 *                  CPU IP6 - Int4: unused      IRQ  6
 *                  CPU IP7 - counter           IRQ  7
 *                  ICU SYSINT1REG bit 0        IRQ  8
 *                  ...
 *                  ICU SYSINT1REG bit 7        IRQ 15
 *                  GPIO 0                      IRQ 40
 *                  ...
 *                  GPIO 31                     IRQ 71
 *                  ICU SYSINT1REG bit 9        IRQ 17
 *                  ...
 *                  ICU SYSINT1REG bit 15       IRQ 23
 *                  ICU SYSINT2REG bit 0        IRQ 24
 *                  ...
 *                  ICU SYSINT2REG bit 15       IRQ 39
 *                  CPU IP0 - software 1        IRQ  0
 * Lowest ---       CPU IP1 - software 2        IRQ  1
 *
 * IRQ 2 is the cascade to the ICU (CPU IP2, Int0)
 *
 * then we just return, if multiple IRQs are pending then we will just take
 * another exception immediately.
 *
 */
		.align	5
		NESTED(vr41xx_handle_irq, PT_SIZE, ra)
		.set	noat
		SAVE_ALL
		CLI
		.set	at
		.set	noreorder

		// Get the pending interrupts
		mfc0	t0,CP0_CAUSE

		// Isolate the allowed ones by anding with irq mask
		mfc0	t2,CP0_STATUS

		andi	t0,0xff00
		and	t0,t2
		andi	t1,t0,0x0800	// check for IRQ 3
		bnez	t1,handle_it
		 li	a0,3
		andi	t1,t0,0xf000	// check for IRQ 4-7
		bnez	t1,1f
		 andi	t1,t0,0x0400	// check for IRQ 2
		bnez	t1,icu_int
		 lui	t3,%hi(VR41XX_SYSINT1REG)
		beqz	t0,go_spurious	// check for IRQ 0-1
		 li	a0,-1
		sll	t0,4
1:		andi	t1,t0,0x1000
		srl	t0,1
		beqz	t1,1b
		 addiu	a0,1
		b	handle_it
		 nop

/*
 *  In the interest of speed, we implement a 3 tier check for which bit
 *  in the ICU level 1 interrupt registers is active.  First, we determine
 *  which 16-bit register (1 or 2).  Second, we narrow it down to which
 *  group of 4 bits, then we get the exact bit.  a0 holds the IRQ number.
 *  Since we always add both 4 and 1 to a0, we initalize it to 5 less than
 *  the lowest IRQ number for each ICU level 1 register
 */
icu_int:	// see delay slot instruction from branch that got us here
		lhu	t0,%lo(VR41XX_SYSINT1REG)(t3)
		lhu	t2,%lo(VR41XX_MSYSINT1REG)(t3)
/* 
 *  OK, this next instruction deserves some explanation.  The timer interrupt
 *  (a few others are similar, but the timer interrupt happens... well... all
 *  the time) comes in on both CPU int1 (IRQ 3) and ICU SYSINT1 bit 2 (IRQ 10),
 *  and since we check IRQ 3 before even loading the value of SYSINT1, it is
 *  possible for a timer interrupt to occur between the time we load CP0_STATUS to
 *  check IRQ 3 and the time we load SYSINT1REG to check IRQ 10.  This will only
 *  happen when we're in the middle of servicing another interrupt, so we just
 *  ignore IRQ 10 (and a few others that behave similarly) and keep looking for
 *  the interrupt that actually caused the exception.  When we re-enable
 *  interrupts, the timer int will immediately re-trigger and be detected as IRQ 3.
 */
		and	t0,0xfffb
		and	t0,t2
		beqz	t0,2f
		 andi	t1,t0,0x01ff	// check for GIU int
		xori	t1,0x0100	// (and no higher priority SYSINT1 int)
		bnez	t1,1f
		 li	a0,3

		// if it is, break out further
#ifdef CONFIG_CPU_VR4181
		lui	t3,%hi(VR41XX_GPINTMSK)
		lhu	t0,%lo(VR41XX_GPINTMSK)(t3)
		lhu	t2,%lo(VR41XX_GPINTSTAT)(t3)
		xori	t0, 0xffff
		and	t0,t2
		bnez	t0,1f
		 li	a0,35
#else
		lhu	t0,%lo(VR41XX_GIUINTLREG)(t3)
		lhu	t2,%lo(VR41XX_MGIUINTLREG)(t3)
		li	a0,35
		and	t0,t2
		bnez	t0,1f
		 lhu	t2,%lo(VR41XX_MGIUINTHREG)(t3)
		lhu	t0,%lo(VR41XX_GIUINTHREG)(t3)
		// load delay stall
		and	t0,t2
		bnez	t0,1f
		 li	a0,51
#endif

go_spurious:	j	spurious_interrupt
		 nop

2:		lhu	t0,%lo(VR41XX_SYSINT2REG)(t3)
		lhu	t2,%lo(VR41XX_MSYSINT2REG)(t3)
#ifdef CONFIG_CPU_VR4181
		and	t0,0xfffe	// see above comment
#else
		and	t0,0xfffa	// see above comment
#endif
		and	t0,t2
		beqz	t0,go_spurious
		 li	a0,19

1:		andi	t1,t0,0x000f
		 addi	a0,4
		beqzl	t1,1b		// the beqzl will cancel the last
		 srl	t0,4		//  shift so we don't lose the bits

1:		andi	t1,t0,0x0001
		 addi	a0,1
		beqz	t1,1b
		 srl	t0,1

handle_it:	jal	do_IRQ
		 move	a1,sp

		j	ret_from_irq
		 nop
		
		END(vr41xx_handle_irq)
