/*
 * linux/arch/mips/vr41xx/kbd.c
 *
 * VR41xx keyboard (KIU) driver
 *
 * Copyright (C) 1999 Bradley D. LaRonde
 * Oct 1999 Hiroshi Kawashima (kawashima@iname.com) adjusted scan lines for MobilPro.
 * Copyright (C) 2000 Michael Klar <wyldfier@iname.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/sched.h>
#include <linux/tty.h>
#include <linux/mm.h>
#include <linux/kbd_kern.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/pm.h>
#include <asm/keyboard.h>
#include <linux/kbd_ll.h>
#include <asm/vr41xx.h>
#include <asm/vr41xx-platdep.h>

void kbd_leds(unsigned char leds)
{
	return;
}

// MFK: do these need to work?  necessary for loadkeys?
int kbd_setkeycode(unsigned int scancode, unsigned int keycode)
{
	return -EINVAL;
}

int kbd_getkeycode(unsigned int scancode)
{
	return -EINVAL;
}

int kbd_translate(unsigned char scancode, unsigned char *keycode,
	char raw_mode)
{
	*keycode = scancode;
	return 1;
}

char kbd_unexpected_up(unsigned char keycode)
{
	printk(KERN_WARNING "vr41xx_kiu: unexpected up, keycode 0x%x\n", keycode);
	return 0x80;
}

#define DEFAULT_REP_WAIT (HZ/4)
#define DEFAULT_REP_RATE (HZ/25)

static int rep_scancode = -1;
static unsigned long rep_wait = DEFAULT_REP_WAIT;
static unsigned long rep_rate = DEFAULT_REP_RATE;
static unsigned long next_reptime;

inline static void handle_cmp_kiudat_bit(unsigned short kiudat, unsigned short cmp_kiudat,
	int scancode, unsigned short bitmask)
{
	int down;

	// did this bit change?
	if ( cmp_kiudat & bitmask ) {
		down = kiudat & bitmask ? 1 : 0;
		if (down) {
			// on press, set up repeat
			rep_scancode = scancode;
			next_reptime = jiffies + rep_wait;
		} else {
			// on release, stop repeating
			if (rep_scancode == scancode)
				rep_scancode = -1;
		}

		// send the scancode and down or up to higher level code
		handle_scancode(scancode, down);
	}
}

inline static void handle_cmp_kiudat_nbl(unsigned short kiudat, unsigned short cmp_kiudat,
	int scancode, unsigned short nblmask)
{
	unsigned short bitmask;

	// common case if for nothing set
	if (cmp_kiudat & nblmask) {
		// handle each bit
		for (bitmask = nblmask & 0x1111 ; bitmask & nblmask; bitmask <<= 1)
			handle_cmp_kiudat_bit(kiudat, cmp_kiudat, scancode++, bitmask);
	}
}

inline static void handle_kiudat(unsigned short kiudat, unsigned short last_kiudat,
	int scancode_base)
{
	unsigned short cmp_kiudat, nblmask;
	int scancode = scancode_base;

	// compare this data with the previous data, setting only bits that changed
	cmp_kiudat = kiudat ^ last_kiudat;

	// nothing changed - most common case
	if ( cmp_kiudat == 0 )
		return;

	for (nblmask = 0x000f; nblmask; nblmask <<= 4) {
		// handle each nibble
		handle_cmp_kiudat_nbl(kiudat, cmp_kiudat, scancode, nblmask);
		scancode += 4;
	}
}

#ifdef CONFIG_CPU_VR4181
#define KIUDAT_REGISTERS 8
#else
#define KIUDAT_REGISTERS 6
#endif

inline static void handle_kiu_data_ready(void)
{
	static unsigned short last_kiudat[KIUDAT_REGISTERS];
	unsigned short thisdat, lastdat;
	volatile unsigned short *kiudatreg = VR41XX_KIUDAT0;
	int i;

	for (i = 0; i < KIUDAT_REGISTERS; i++) {
		lastdat = last_kiudat[i];
#ifdef CONFIG_CPU_VR4181
// not clear whether this is right:
		thisdat = last_kiudat[i] = ~(*(kiudatreg++));
#else
		thisdat = last_kiudat[i] = *(kiudatreg++);
#endif
		handle_kiudat(thisdat, lastdat, i * 16);
	}
	if ((rep_scancode >= 0) && (time_after_eq(jiffies, next_reptime))) {
		next_reptime = jiffies + rep_rate;
		handle_scancode(rep_scancode, 1);
	}
}

void kbd_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned short kiuint;

	// get keyboard interrupt status
	kiuint = *VR41XX_KIUINT;
	barrier();

	// ack them, but only the ones that were read.  if we just ack
	// all of them, we run the risk of ignoring interrupts that
	// happen between these 2 lines
	*VR41XX_KIUINT = kiuint;

	// keyboard data ready?
	if ( kiuint & VR41XX_KIUINT_KDATRDY ) {
		handle_kiu_data_ready();
	}

	// lost keyboard data?
	if ( kiuint & VR41XX_KIUINT_KDATLOST ) {
		printk(KERN_DEBUG "vr41xx_kiu: keyboard data lost\n");
	}
}

#ifdef CONFIG_PM
static int pm_kbd_request(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	switch (rqst) {
	case PM_SUSPEND:
// MFK: there should be some kind of enable/disable for this (ie, "enable
// wake-up on this device"), but user interface for that not worked out yet.
		*VR41XX_MKIUINTREG = VR41XX_KIUINT_SCANINT;
		break;
	case PM_RESUME:
		*VR41XX_MKIUINTREG = VR41XX_KIUINT_KDATRDY | VR41XX_KIUINT_KDATLOST;
		break;
	}
	return 0;
}
#endif

void __init kbd_init_hw(void)
{
	printk(KERN_INFO "Keyboard driver version 0.2a for VR41XX KIU\n");

#ifndef CONFIG_CPU_VR4181
	// turn on the KIU clock
	*VR41XX_CMUCLKMSK |= 0x0008;

	// reset the KIU
	*VR41XX_KIURST = 0x0001;

	barrier();

	// number of scanlines to use varies by device, but most use all 12
#if (KBD_SCANLINES == 8)
	*VR41XX_SCANLINE = 0x0002;
#elif (KBD_SCANLINES == 10)
	*VR41XX_SCANLINE = 0x0001;
#elif (KBD_SCANLINES == 12)
	*VR41XX_SCANLINE = 0x0000;
#else
#error "Illegal number of scanlines defined in KBD_SCANLINES"
#endif
#endif

	// enable scan, auto scan, auto stop, and one stop rep
	*VR41XX_KIUSCANREP = VR41XX_KIUSCANREP_KEYEN |
		 VR41XX_KIUSCANREP_ATSCAN | VR41XX_KIUSCANREP_ATSTP | (1 << 4);

	// set stabilization, off, and wait times in 30us increments
	// this uses a short (but not the shortest) setting for each
#ifdef CONFIG_CPU_VR4181
	// VR4181 off time is fixed at 1
	*VR41XX_KIUWKS = (9-1) | ((17-1) << 10);
#else
	*VR41XX_KIUWKS = (9-1) | ((9-1) << 5) | ((9-1) << 10);
#endif

	// set wait time between scan cycles in 30us increments
	// setting this too short can result in lost keyboard data
	// this wait is on top of the stabilization, off, and wait times above,
	// which are per scan line, so total cycle time is about 1/50 sec
	*VR41XX_KIUWKI = (20000 / 30) - KBD_SCANLINES * 27;

	barrier();

	// clear any pending interrupts
	*VR41XX_KIUINT = 0x0007; 

	barrier();

	// install the interrupt handler
	request_irq(VR41XX_IRQ_KIU, kbd_irq_handler, 0, "keyboard", NULL);

	// unmask the keyboard interrupts that we are interested in
	*VR41XX_MKIUINTREG = VR41XX_KIUINT_KDATRDY | VR41XX_KIUINT_KDATLOST;

#ifdef CONFIG_PM
	pm_register(PM_SYS_DEV, PM_SYS_KBC, pm_kbd_request);
#endif
}
